#include <kernel/fs.h>
#include <errno.h>
#include <kernel/lib_k.h>
//----------------------------------------------------------------------
zno_t
inode_zone (inode_t *inode, zno_t fzone, int write)
{
    unsigned int indirect_zones;
    unsigned int allocated_zone;
    zno_t        zone_table[INODE_MAX_INDIRECT_ZONES];
    char         buffer[SB_MAX_ZONE_SIZE];
    unsigned int i0;            // Single indirect index.     
    unsigned int i1;            // Double indirect first index.
    unsigned int i2;            // Double indirect second index.
    int          status;
    zno_t        zone_second;   // Second level table zone.
    //
    // Check to have a valid inode.
    //
    if (inode == NULL)
      {
        errset (EINVAL);
        return ((zno_t) -1);
      }
    //
    // Calculate how many indirect zone numbers are stored inside
    // a zone: it depends on the zone size.
    //
    indirect_zones  = inode->sb->blksize / 2;
    //
    // Convert file-zone number into a zone number.
    //
    if (fzone < 7)
      {
        //
        // 0 <= fzone <= 6
        // The zone number is inside the direct zone references.
        // Verify to have such zone.
        //
        if (inode->direct[fzone] == 0)
          {
            //
            // There is not such zone, but we do not consider
            // it an error, because a file can be not contiguous.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // Must be allocated.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // Cannot allocate the zone. The variable `errno' is
                // set by `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone is allocated: clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone is allocated and cleared: save the inode.
            //
            inode->direct[fzone] = allocated_zone;
            inode->changed = 1;
            status = inode_save (inode);
            if (status != 0)
              {
                //
                // Cannot save the inode. The variable `errno' is
                // set `inode_save()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // The zone is there: return it.
        //
        return (inode->direct[fzone]);
      }
    if (fzone < 7 + indirect_zones)
      {
        //
        // 7 <= fzone <= (6 + indirect_zones)
        // The zone number is inside the single indirect zone
        // references: verify to have the indirect zone table.
        //
        if (inode->indirect1 == 0)
          {
            //
            // There is not such zone, but it is not an error.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // The first level of indirection must be initialized.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // Cannot allocate the zone for the indirection table:
                // this is an error and the `errno' value is produced
                // by `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone for the indirection table is allocated:
            // clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // The indirection table zone is allocated and cleared:
            // save the inode.
            //
            inode->indirect1 = allocated_zone;
            inode->changed = 1;
            status = inode_save (inode);
            if (status != 0)
              {
                //
                // Cannot save the inode. This is an error and the value
                // for `errno' is produced by `inode_save()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // An indirect table is present inside the file system:
        // load it.
        //
        status = zone_read (inode->sb, inode->indirect1, zone_table);
        if (status != 0)
          {
            //
            // Cannot load the indirect table. This is an error and the
            // value for `errno' is assigned by function `zone_read()'.
            //
            return ((zno_t) -1);
          }
        //
        // The indirect table was read. Calculate the index inside
        // the table, for the requested zone.
        //      
        i0 = (fzone - 7);
        //
        // Check if the zone is to be allocated.
        //
        if (zone_table[i0] == 0)
          {
            //
            // There is not such zone, but it is not an error.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // The zone must be allocated.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // There is no space for the zone allocation. The
                // variable `errno' is already updated by
                // `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone is allocated: clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone is allocated and cleared: update the indirect
            // zone table an save it. The inode is not modified,
            // because the indirect table is outside.
            //
            zone_table[i0] = allocated_zone;
            status = zone_write (inode->sb, inode->indirect1,
                                 zone_table);
            if (status != 0)
              {
                //
                // Cannot save the zone. The variable `errno' is already
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // The zone is allocated.
        //
        return (zone_table[i0]);
      }
    else
      {
        //
        // (7 + indirect_zones) <= fzone
        // The zone number is inside the double indirect zone
        // references.
        // Verify to have the first level of second indirection.
        //
        if (inode->indirect2 == 0)
          {
            //
            // There is not such zone, but it is not an error.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // The first level of second indirection must be
            // initialized.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // Cannot allocate the zone. The variable `errno' is
                // set by `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone for the indirection table is allocated:
            // clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone for the indirection table is allocated and
            // cleared: save the inode.
            //
            inode->indirect2 = allocated_zone;
            inode->changed = 1;
            status = inode_save (inode);
            if (status != 0)
              {
                //
                // Cannot save the inode. The variable `errno' is
                // set by `inode_save()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // The first level of second indirection is present:
        // Read the second indirect table.
        //
        status = zone_read (inode->sb, inode->indirect2, zone_table);
        if (status != 0)
          {
            //
            // Cannot read the second indirect table. The variable
            // `errno' is set by `zone_read()'.
            //
            return ((zno_t) -1);
          }
        //
        // The first double indirect table was read: calculate
        // indexes inside first and second level of table.
        //
        fzone -= 7;
        fzone -= indirect_zones;
        i1     = fzone / indirect_zones;
        i2     = fzone % indirect_zones;
        //
        // Verify to have a second level.
        //
        if (zone_table[i1] == 0)
          {
            //
            // There is not such zone, but it is not an error.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // The second level must be initialized.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // Cannot allocate the zone. The variable `errno' is set
                // by `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone for the indirection table is allocated:
            // clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // Update the first level index and save it.
            //
            zone_table[i1] = allocated_zone;
            status = zone_write (inode->sb, inode->indirect2,
                                 zone_table);
            if (status != 0)
              {
                //
                // Cannot write the zone. The variable `errno' is set
                // by `zone_write()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // The second level can be read, overwriting the array
        // `zone_table[]'. The zone number for the second level
        // indirection table is saved inside `zone_second', before
        // overwriting the array.
        //
        zone_second = zone_table[i1];
        status = zone_read (inode->sb, zone_second, zone_table);
        if (status != 0)
          {
            //
            // Cannot read the second level indirect table. The variable
            // `errno' is set by `zone_read()'.
            //
            return ((zno_t) -1);
          }
        //
        // The second level was read and `zone_table[]' is now
        // such second one: check if the zone is to be allocated.
        //
        if (zone_table[i2] == 0)
          {
            //
            // There is not such zone, but it is not an error.
            //
            if (!write)
              {
                return ((zno_t) 0);
              }
            //
            // Must be allocated.
            //
            allocated_zone = zone_alloc (inode->sb);
            if (allocated_zone == 0)
              {
                //
                // Cannot allocate the zone. The variable `errno' is set
                // by `zone_alloc()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone is allocated: clear the zone and save.
            //
            memset (buffer, 0, SB_MAX_ZONE_SIZE);
            status = zone_write (inode->sb, allocated_zone, buffer);
            if (status < 0)
              {
                //
                // Cannot overwrite the zone. The variable `errno' is
                // set by `zone_write()'.
                //
                return ((zno_t) -1);
              }
            //
            // The zone was allocated and cleared: update the indirect
            // zone table an save it. The inode is not modified, because
            // the indirect table is outside.
            //
            zone_table[i2] = allocated_zone;
            status = zone_write (inode->sb, zone_second, zone_table);
            if (status != 0)
              {
                //
                // Cannot write the zone. The variable `errno' is set
                // by `zone_write()'.
                //
                return ((zno_t) -1);
              }
          }
        //
        // The zone is there: return the zone number.
        //
        return (zone_table[i2]);
      }
}
