#include <kernel/fs.h>
#include <errno.h>
#include <kernel/proc.h>
#include <kernel/lib_k.h>
//----------------------------------------------------------------------
#define DIRECTORY_BUFFER_SIZE (SB_MAX_ZONE_SIZE/16)
//----------------------------------------------------------------------
inode_t *
path_inode (pid_t pid, const char *path)
{
    proc_t     *ps;
    inode_t    *inode;
    dev_t       device;
    char        full_path[PATH_MAX];
    char       *name;
    char       *next;
    directory_t dir[DIRECTORY_BUFFER_SIZE];
    char        dir_name[NAME_MAX+1];
    off_t       offset_dir;
    ssize_t     size_read;
    size_t      dir_size_read;
    ssize_t     size_to_read;
    int         comp;
    int         d;              // Directory index;
    int         status;         // inode_check() return status.
    //
    // Get process.
    //
    ps = proc_reference (pid);
    //
    // Arrange to get a packed full path name.
    //
    path_full (path, ps->path_cwd, full_path);
    //
    // Get the root file system inode.
    //
    inode = inode_get ((dev_t) 0, 1);
    if (inode == NULL)
      {
        errset (errno);
        return (NULL);
      }
    //
    // Save the device number.
    //
    device = inode->sb->device;
    //
    // Variable `inode' already points to the root file system inode:
    // It must be a directory!
    //
    status = inode_check (inode, S_IFDIR, 1, ps->euid, ps->egid);
    if (status != 0)
      {
        //
        // Variable `errno' should be set by inode_check().
        //
        errset (errno);
        inode_put (inode);
        return (NULL);
      }
    //
    // Initialize string scan: find the first path token, after the
    // first `/'.
    //
    name = strtok (full_path, "/");
    //
    // If the original full path is just `/' the variable `name'
    // appears as a null pointer, and the variable `inode' is already
    // what we are looking for.
    //
    if (name == NULL)
      {
        return (inode);
      }
    //
    // There is at least a name after `/' inside the original full
    // path. A scan is going to start: the original value for variable
    // `inode' is a pointer to the root directory inode.
    //
    for (;;)
      {
        //
        // Find next token.
        //
        next = strtok (NULL, "/");
        //
        // Read the directory from the current inode.
        //
        for (offset_dir=0; ; offset_dir += size_read)
          {
            size_to_read = DIRECTORY_BUFFER_SIZE;
            //
            if ((offset_dir + size_to_read) > inode->size)
              {
                size_to_read = inode->size - offset_dir;
              }
            //
            size_read = inode_file_read (inode, offset_dir, dir,
                                         size_to_read, NULL);
            //
            // The size read must be a multiple of 16.
            //
            size_read = ((size_read / 16) * 16);
            //
            // Check anyway if it is zero.
            //
            if (size_read == 0)
              {
                //
                // The directory is ended: release the inode and return.
                //
                inode_put (inode);
                errset (ENOENT);           // No such file or directory.
                return (NULL);
              }
            //
            // Calculate how many directory items we have read.
            //
            dir_size_read = size_read / 16;
            //
            // Scan the directory to find the current name.
            //
            for (d = 0; d < dir_size_read; d++)
              {
                //
                // Ensure to have a null terminated string for
                // the name found.
                //
                memcpy (dir_name, dir[d].name, (size_t) NAME_MAX);
                dir_name[NAME_MAX] = 0;
                //
                comp = strcmp (name, dir_name);
                if (comp == 0 && dir[d].ino != 0)
                  {
                    //
                    // Found the name and verified that it has a link to
                    // a inode. Now release the directory inode.
                    //
                    inode_put (inode);
                    //
                    // Get next inode and break the loop.
                    //
                    inode = inode_get (device, dir[d].ino);
                    break;
                  }
              }
            //
            // If index `d' is in a valid range, the name was found.
            //
            if (d < dir_size_read)
              {
                //
                // The name was found.
                //
                break;
              }
          }
        //
        // If the function is still working, a file or a directory
        // was found: see if there is another name after this one
        // to look for. If there isn't, just break the loop.
        //
        if (next == NULL)
          {
            //
            // As no other tokens are to be found, break the loop.
            //
            break;
          }
        //
        // As there is another name after the current one,
        // the current file must be a directory.
        //
        status = inode_check (inode, S_IFDIR, 1, ps->euid, ps->egid);
        if (status != 0)
          {
            //
            // Variable `errno' is set by `inode_check()'.
            //
            errset (errno);
            inode_put (inode);
            return (NULL);
         }
        //
        // The inode is a directory and the user has the necessary
        // permissions: check if it is a mount point and go to the
        // new device root directory if necessary.
        //
        if (inode->sb_attached != NULL)
          {
            //
            // Must find the root directory for the new device, and
            // then go to that inode.
            //
            device = inode->sb_attached->device;
            inode_put (inode);
            inode = inode_get (device, 1);
            status = inode_check (inode, S_IFDIR, 1,
                                  ps->euid, ps->egid);
            if (status != 0)
              {
                inode_put (inode);
                return (NULL);
              }
          }
        //
        // As a directory was found, and another token follows it,
        // must continue the token scan.
        //
        name = next;
      }
    //
    // Current inode found is the file represented by the requested
    // path.
    //
    return (inode);
}
