#include <kernel/proc.h>
#include <errno.h>
#include <fcntl.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
extern uint32_t proc_stack_pointer;
//----------------------------------------------------------------------
pid_t
s_fork (pid_t ppid)
{
  pid_t           pid;
  pid_t           zombie;
  addr_t          allocated_text = 0;
  addr_t          allocated_data = 0;
  addr_t          addr_stack_pointer = 0;
  int             fdn;
  uint16_t        segment_descriptor;
  int             sig;
  //
  // Find a free PID.
  //
  for (pid = 1; pid < PROCESS_MAX; pid++)
    {
      if (proc_table[pid].status == PROC_EMPTY)
	{
	  break;
	}
    }
  if (pid >= PROCESS_MAX)
    {
      //
      // There is no free pid.
      //
      errset (ENOMEM);		// Not enough space.
      return (-1);
    }
  //
  // Before allocating a new process, must check if there are some
  // zombie slots, still with original segment data: should reset
  // them now!
  //
  for (zombie = 1; zombie < PROCESS_MAX; zombie++)
    {
      if (proc_table[zombie].status == PROC_ZOMBIE
	  && (proc_table[zombie].address_text != 0
	      || proc_table[zombie].domain_text != 0))
	{
	  proc_table[zombie].address_text = (addr_t) 0;
	  proc_table[zombie].domain_text = (size_t) 0;
	  proc_table[zombie].address_data = (addr_t) 0;
	  proc_table[zombie].domain_data = (size_t) 0;
	  proc_table[zombie].domain_stack = (size_t) 0;
	  proc_table[zombie].extra_data = (size_t) 0;
	  proc_table[zombie].sp = 0;
	}
    }
  //
  // Allocate memory for text, if text and data are inside
  // the same address space.
  //
  if (proc_table[ppid].domain_data == 0)
    {
      allocated_text = mb_alloc_size (proc_table[ppid].domain_text
				      + proc_table[ppid].extra_data);
      //
      if (allocated_text == 0)
	{
	  errset (ENOMEM);	// Not enough space.
	  return ((pid_t) - 1);
	}
      //
      if (DEBUG)
	{
	  k_printf ("%s:%i:mb_alloc_size(%zi)",
		    __FILE__, __LINE__,
		    (proc_table[ppid].domain_text
		     + proc_table[ppid].extra_data));
	}
    }
  //
  // Allocate memory for data, if necessary.
  //
  if (proc_table[ppid].domain_data > 0)
    {
      allocated_data = mb_alloc_size (proc_table[ppid].domain_data
				      + proc_table[ppid].extra_data);
      //
      if (allocated_data == 0)
	{
	  //
	  // Please note that, if we are here, no memory
	  // for the text was allocated!
	  //
	  errset (ENOMEM);	// Not enough space.
	  return ((pid_t) - 1);
	}
      //
      if (DEBUG)
	{
	  k_printf ("%s:%i:mb_alloc_size(%zi)",
		    __FILE__, __LINE__,
		    (proc_table[ppid].domain_data
		     + proc_table[ppid].extra_data));
	}
    }
  //
  // Copy the process text and, data in memory: if
  // size is zero, no copy is made. But the text is
  // copied only if text and data live together.
  //
  if (proc_table[ppid].domain_data == 0)
    {
      memcpy ((void *) allocated_text,
	      (void *) proc_table[ppid].address_text,
	      (size_t) (proc_table[ppid].domain_text
			+ proc_table[ppid].extra_data));
    }
  else
    {
      memcpy ((void *) allocated_data,
	      (void *) proc_table[ppid].address_data,
	      (size_t) (proc_table[ppid].domain_data
			+ proc_table[ppid].extra_data));
    }
  //
  // Allocate the new PID inside the `proc_table[]'.
  //
  proc_table[pid].ppid = ppid;
  proc_table[pid].pgrp = proc_table[ppid].pgrp;
  proc_table[pid].uid  = proc_table[ppid].uid;
  proc_table[pid].euid = proc_table[ppid].euid;
  proc_table[pid].suid = proc_table[ppid].suid;
  proc_table[pid].gid  = proc_table[ppid].gid;
  proc_table[pid].egid = proc_table[ppid].egid;
  proc_table[pid].sgid = proc_table[ppid].sgid;
  proc_table[pid].device_tty = proc_table[ppid].device_tty;
  proc_table[pid].sig_status = 0;
  proc_table[pid].sig_ignore = 0;
  //
  for (sig = 0; sig < MAX_SIGNALS; sig++)
    {
      proc_table[pid].sig_handler[sig]
	= proc_table[ppid].sig_handler[sig];
    }
  //
  proc_table[pid].usage = 0;
  proc_table[pid].status = PROC_CREATED;
  proc_table[pid].wakeup_events = 0;
  proc_table[pid].wakeup_signal = 0;
  proc_table[pid].wakeup_timer = 0;
  //
  if (proc_table[ppid].domain_data != 0)
    {
      proc_table[pid].address_text = proc_table[ppid].address_text;
    }
  else
    {
      proc_table[pid].address_text = allocated_text;
    }
  proc_table[pid].domain_text = proc_table[ppid].domain_text;
  proc_table[pid].address_data = allocated_data;
  proc_table[pid].domain_data = proc_table[ppid].domain_data;
  proc_table[pid].domain_stack = proc_table[ppid].domain_stack;
  proc_table[pid].extra_data = proc_table[ppid].extra_data;
  proc_table[pid].sp = proc_stack_pointer;
  proc_table[pid].ret = 0;
  proc_table[pid].inode_cwd = proc_table[ppid].inode_cwd;
  proc_table[pid].umask = proc_table[ppid].umask;
  strncpy (proc_table[pid].name, proc_table[ppid].name, PATH_MAX);
  strncpy (proc_table[pid].path_cwd,
	   proc_table[ppid].path_cwd, PATH_MAX);
  //
  // Update process TEXT segment inside the GDT table.
  //
  gdt_segment (gdt_pid_to_segment_text (pid),
	       (uint32_t) proc_table[pid].address_text,
	       (uint32_t) (proc_table[pid].domain_text / 4096),
	       1, 1, 0);
  //
  // Update process DATA segment inside the GDT table.
  //
  if (proc_table[pid].domain_data > 0)
    {
      gdt_segment (gdt_pid_to_segment_data (pid),
		   (uint32_t) proc_table[pid].address_data,
		   (uint32_t) ((proc_table[pid].domain_data
				+ proc_table[pid].extra_data) / 4096),
		   1, 0, 0);
    }
  else
    {
      gdt_segment (gdt_pid_to_segment_data (pid),
		   (uint32_t) proc_table[pid].address_text,
		   (uint32_t) ((proc_table[pid].domain_text
				+ proc_table[pid].extra_data) / 4096),
		   1, 0, 0);
    }
  //
  //------------------------------------------------------------------
  // Might reload the GDT table, but it is not necessarily.
  // Anyway, if you do it, nothing change. :-)
  //
  // gdt_load  (&gdt_register);
  //------------------------------------------------------------------
  //
  // Increase inode references for the working directory.
  //
  proc_table[pid].inode_cwd->references++;
  //
  // Duplicate valid file descriptors.
  //
  for (fdn = 0; fdn < OPEN_MAX; fdn++)
    {
      if (proc_table[ppid].fd[fdn].file != NULL
	  && (proc_table[ppid].fd[fdn].file->inode != NULL
	      || proc_table[ppid].fd[fdn].file->sock != NULL))
	{
	  //
	  // Copy to the forked process.
	  //
	  proc_table[pid].fd[fdn].fl_flags
	    = proc_table[ppid].fd[fdn].fl_flags;
	  proc_table[pid].fd[fdn].fd_flags
	    = proc_table[ppid].fd[fdn].fd_flags;
	  proc_table[pid].fd[fdn].file = proc_table[ppid].fd[fdn].file;
	  //
	  // Increment file reference.
	  //
	  proc_table[ppid].fd[fdn].file->references++;
	  //
	  // Check if it is a pipe and increment specific
	  // read/write reference counters inside the inode.
	  //
	  if (proc_table[ppid].fd[fdn].file->inode != NULL
	      && S_ISFIFO (proc_table[ppid].fd[fdn].file->inode->mode))
	    {
	      if (proc_table[ppid].fd[fdn].fl_flags & O_RDONLY)
		{
		  proc_table[ppid].fd[fdn].file->inode->pipe_ref_read++;
		}
	      //
	      if (proc_table[ppid].fd[fdn].fl_flags & O_WRONLY)
		{
		  proc_table[ppid].fd[fdn].file->inode->
		    pipe_ref_write++;
		}
	    }
	}
    }
  //
  // Change segment descriptor values inside the stack,
  // for: DS==ES==FS==GS.
  //
  // First calculate the absolute stack section address, from the
  // kernel point of view.
  //
  if (allocated_data > 0)
    {
      addr_stack_pointer = allocated_data;
    }
  else
    {
      addr_stack_pointer = allocated_text;
    }
  //
  // Then calculate the effective new stack pointer.
  //
  addr_stack_pointer += proc_table[pid].sp;
  //
  // Then calculate the segment descriptor, to be written
  // inside the new process stack.
  //
  segment_descriptor = (gdt_pid_to_segment_data (pid) * 8);
  //
  // Then copy inside the stack the new values for data segments.
  //
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
	  (addr_stack_pointer + 28),
	  &segment_descriptor, (sizeof segment_descriptor), NULL);
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
	  (addr_stack_pointer + 32),
	  &segment_descriptor, (sizeof segment_descriptor), NULL);
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
	  (addr_stack_pointer + 36),
	  &segment_descriptor, (sizeof segment_descriptor), NULL);
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
	  (addr_stack_pointer + 40),
	  &segment_descriptor, (sizeof segment_descriptor), NULL);
  //
  // Change segment descriptor value inside the stack for CS,
  // if so is necessary.
  //
  segment_descriptor = (gdt_pid_to_segment_text (pid) * 8);
  //
  dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
	  (addr_stack_pointer + 48),
	  &segment_descriptor, (sizeof segment_descriptor), NULL);
  //
  // Set it ready.
  //
  proc_table[pid].status = PROC_READY;
  //
  // Return the new PID.
  //
  return (pid);
}
