#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <errno.h>
#include <fcntl.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
ssize_t
s_read (pid_t pid, int fdn, void *buffer, size_t count)
{
  fd_t *fd;
  ssize_t size_read;
  int eof = 0;
  //
  // Get file descriptor.
  //
  fd = fd_reference (pid, &fdn);
  if (fd == NULL || fd->file == NULL
      || (fd->file->inode == NULL && fd->file->sock == NULL))
    {
      errset (EBADF);           // Bad file descriptor.
      return ((ssize_t) - 1);
    }
  //
  // Check if it is opened for read.
  //
  if (!(fd->file->oflags & O_RDONLY))
    {
      //
      // The file is not opened for read.
      //
      errset (EINVAL);          // Invalid argument.
      return ((ssize_t) - 1);
    }
  //
  // Check the kind of file to be read and read it.
  //
  if (fd->file->sock != NULL)
    {
      //
      // Read from the socket and return.
      //
      return (s_recvfrom (pid, fdn, buffer, count, 0, NULL, NULL));
    }
  else if (S_ISBLK (fd->file->inode->mode)
           || S_ISCHR (fd->file->inode->mode))
    {
      //
      // A device is to be read.
      //
      size_read = dev_io (pid, (dev_t) fd->file->inode->direct[0],
                          DEV_READ, fd->file->offset, buffer, count,
                          &eof);
      if (size_read < 0 && (errno == EAGAIN || errno == EWOULDBLOCK))
        {
          if (fd->fl_flags & O_NONBLOCK)
            {
              //
              // Non blocking null read.
              //
              ;
            }
          else
            {
              //
              // Null read: put the process to sleep.
              //
              proc_table[pid].status = PROC_SLEEPING;
              proc_table[pid].ret = 0;
              proc_table[pid].wakeup_events = WAKEUP_EVENT_DEV_READ;
              proc_table[pid].wakeup_dev = fd->file->inode->direct[0];
              if (DEBUG)
                {
                  k_printf ("[%s] PID %i goes to sleep "
                            "waiting to read from a device.\n",
                            __FILE__, pid);
                }
            }
        }
    }
  else if (S_ISREG (fd->file->inode->mode))
    {
      //
      // A regular file is to be read.
      //
      size_read = inode_file_read (fd->file->inode, fd->file->offset,
                                   buffer, count, &eof);
    }
  else if (S_ISDIR (fd->file->inode->mode))
    {
      //
      // A directory, is to be read.
      //
      size_read = inode_file_read (fd->file->inode, fd->file->offset,
                                   buffer, count, &eof);
    }
  else if (S_ISFIFO (fd->file->inode->mode))
    {
      //
      // A pipe, is to be read.
      //
      size_read = inode_pipe_read (fd->file->inode, buffer, count,
                                   &eof);
      //
      if (size_read == 0)
        {
          //
          // Check what to do.
          //
          if (fd->file->inode->pipe_ref_write == 0)
            {
              //
              // EOF, if it is a valid pointer, is already
              // set by `inode_pipe_read()', if is time to
              // set it.
              //
              // Wake up processes waiting to write.
              //
              proc_wakeup_pipe_write (fd->file->inode);
              //
              return (size_read);
            }
          else
            {
              //
              // Go to sleep.
              //
              proc_table[pid].status = PROC_SLEEPING;
              proc_table[pid].ret = 0;
              proc_table[pid].wakeup_inode = fd->file->inode;
              proc_table[pid].wakeup_events = WAKEUP_EVENT_PIPE_READ;
              if (DEBUG)
                {
                  k_printf ("[%s] PID %i goes to sleep "
                            "waiting to read from a pipe.\n",
                            __FILE__, pid);
                }
            }
        }
      else
        {
          //
          // Wake up processes waiting to write.
          //
          proc_wakeup_pipe_write (fd->file->inode);
        }
    }
  else
    {
      //
      // Unsupported file type.
      //
      errset (E_FILE_TYPE_UNSUPPORTED); //File type unsupported.
      return ((ssize_t) - 1);
    }
  //
  // Update the file descriptor internal offset, if there is an inode.
  //
  if (fd->file->inode != NULL && size_read > 0)
    {
      fd->file->offset += size_read;
    }
  //
  // Return the size read, even if it is an error. Please notice
  // that a size of zero might be related to an end of file, or
  // just that the read should be retried. For the latter case,
  // -1 is returned with error EAGAIN, so that the function
  // `read()' can retry.
  //
  if (size_read == 0 && !eof)
    {
      errset (EAGAIN);
      return (-1);
    }
  else
    {
      return (size_read);
    }
}
