#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
//----------------------------------------------------------------------
int
s_socket (pid_t pid, int family, int type, int protocol)
{
  fd_t    *fd;
  int      sfdn;
  file_t  *file;
  sock_t  *sock;
  //
  // Check supported family type.
  //
  if (family != AF_INET)
    {
      errset (EAFNOSUPPORT);
      return (-1);
    }
  //
  // Check supported communication type.
  //
  if (type == SOCK_RAW || type == SOCK_DGRAM || type == SOCK_STREAM)
    {
      //
      // Ok.
      //
      ;
    }
  else
    {
      errset (EPROTONOSUPPORT);
      return (-1);
    }
  //
  // Check supported protocol type.
  //
  if (protocol == IPPROTO_ICMP
      || protocol == IPPROTO_UDP
      || protocol == IPPROTO_TCP)
    {
      //
      // Ok.
      //
      ;
    }
  else
    {
      errset (EPROTONOSUPPORT);
      return (-1);
    }
  //
  // If it is a raw socket, must be a privileged process.
  //
  if (type == SOCK_RAW && proc_table[pid].euid != 0)
    {
      errset (EACCES);
      return (-1);
    }
  //
  // Find a free slot inside the sock_table[].
  //
  sock = sock_reference (-1);
  if (sock == NULL)
    {
      errset (ENFILE);
      return (-1);
    }
  //
  // Find a free slot inside the file table.
  //
  file = file_reference (-1);
  if (file == NULL)
    {
      errset (ENFILE);          // Too many files open in system.
      return (-1);
    }
  //
  // Find a free slot inside the file descriptor table.
  // Variable `sfdn' will be modified by the call to `fd_reference()'.
  //
  sfdn = -1;
  fd = fd_reference (pid, &sfdn);
  if (fd == NULL)
    {
      errset (EMFILE);          // Too many open files.
      return (-1);
    }
  //
  // socket, system file and file descriptor ready: reset and put data
  // inside them. Please notice that the tcp.listen_queue[] array is
  // reset with all 0xFF, because zero is a valid file descriptor
  // number.
  //
  memset (sock, 0, sizeof (sock_t));
  sock->active      = 1;
  sock->family      = family;
  sock->type        = type;
  sock->protocol    = protocol;
  sock->lport       = 0;
  sock->laddr       = 0;
  sock->rport       = 0;
  sock->raddr       = 0;
  memset (sock->read.clock, 0x00, sizeof (sock->read.clock));
  memset (sock->tcp.listen_queue, 0xFF,
          sizeof (sock->tcp.listen_queue));
  //
  file->references = 1;
  file->oflags     = O_RDWR;
  file->inode      = NULL;
  file->sock       = sock;
  file->offset     = 0;
  //
  fd->fl_flags     = (O_RDWR | O_APPEND);
  fd->fd_flags     = 0;
  fd->file         = file;
  //
  // Return the file descriptor.
  //
  return (sfdn);
}
