#include <kernel/fs.h>
#include <errno.h>
#include <kernel/proc.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------------------
int
s_umount (pid_t pid, const char *path_mnt)
{
  proc_t         *ps;
  dev_t           device;	// Device to mount.
  inode_t        *inode_mount_point;	// Original mount point.
  inode_t        *inode;	// Inode table.
  int             i;		// Inode table index.
  //
  // Get process.
  //
  ps = proc_reference (pid);
  //
  // Verify to be the super user.
  //
  if (ps->euid != 0)
    {
      errset (EPERM);		// Operation not permitted.
      return (-1);
    }
  //
  // Get the directory mount point.
  //
  inode_mount_point = path_inode (pid, path_mnt);
  if (inode_mount_point == NULL)
    {
      errset (ENOENT);		// No such file or directory.
      return (-1);
    }
  //
  // Verify that the path is a directory.
  //
  if (!S_ISDIR (inode_mount_point->mode))
    {
      inode_put (inode_mount_point);
      errset (ENOTDIR);		// Not a directory.
      return (-1);
    }
  //
  // Verify that there is something attached.
  //
  device = inode_mount_point->sb_attached->device;
  if (device == 0)
    {
      //
      // There is nothing to unmount.
      //
      inode_put (inode_mount_point);
      errset (E_NOT_MOUNTED);	// Not mounted.
      return (-1);
    }
  //
  // Are there exactly two internal references? Let's explain:
  // the directory that act as mount point, should have one reference
  // because it is mounting something and another because it was just
  // opened again, a few lines above. If there are more references
  // it is wrong; if there are less, it is also wrong at this point.
  //
  if (inode_mount_point->references != 2)
    {
      inode_put (inode_mount_point);
      errset (EUNKNOWN);	// Unknown error.
      return (-1);
    }
  //
  // All data is available: find if there are open file inside
  // the file system to unmount. But first load the inode table
  // pointer.
  //
  inode = inode_reference ((dev_t) 0, (ino_t) 0);
  if (inode == NULL)
    {
      //
      // This error should not happen.
      //
      inode_put (inode_mount_point);
      errset (EUNKNOWN);	// Unknown error.
      return (-1);
    }
  //
  // Scan the inode table.
  //
  for (i = 0; i < INODE_MAX_SLOTS; i++)
    {
      if ((inode[i].sb == inode_mount_point->sb_attached)
          && (inode[i].references > 0))
	{
	  //
	  // At least one file is open inside the super block to
	  // release: cannot unmount.
	  //
	  inode_put (inode_mount_point);
	  errset (EBUSY);	// Device or resource busy.
	  return (-1);
	}
    }
  //
  // Can unmount: save and remove the super block memory;
  // clear the mount point reference and put inode.
  //
  inode_mount_point->sb_attached->changed = 1;
  sb_save (inode_mount_point->sb_attached);
  //
  inode_mount_point->sb_attached->device = 0;
  inode_mount_point->sb_attached->inode_mounted_on = NULL;
  inode_mount_point->sb_attached->blksize = 0;
  inode_mount_point->sb_attached->options = 0;
  //
  inode_mount_point->sb_attached = NULL;
  inode_mount_point->references = 0;
  inode_put (inode_mount_point);
  //
  inode_put (inode_mount_point);
  //
  return (0);
}
