#include <kernel/net.h>
#include <kernel/net/arp.h>
#include <kernel/net/icmp.h>
#include <kernel/net/ip.h>
#include <sys/os32.h>
#include <kernel/lib_k.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/udp.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
ip_rx (int n, int f)
{
  struct iphdr    *header;
  net_ip_packet_t *packet;
  uint16_t         checksum;
  size_t           size_header;
  ip_t            *ip_table_item;
  struct udphdr   *udp;
  int              i;
  int              j;           // Net table index.
  int              s;           // Socket table index.
  //
  //
  //
  if (n >= NET_MAX_DEVICES || n < 0)
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  if (f >= NET_MAX_BUFFERS || f < 0)
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  //
  // Get the packet link.
  //
  if (net_table[n].type & NET_DEV_LOOP)
    {
      packet = (net_ip_packet_t *)
               &net_table[n].loopback.buffer[f].packet;
    }
  else if (net_table[n].type & NET_DEV_ETH)
    {
      //
      // It is Ethernet, but must also have an IP packet inside!
      //
      if (ntohs (net_table[n].ethernet.buffer[f].frame.header.type)
          != NET_PROT_IP)
        {
          errset (EINVAL);          // Invalid argument.
          return (-1);
        }
      packet = (net_ip_packet_t *)
               &net_table[n].ethernet.buffer[f].frame.packet;
    }
  else
    {
      errset (EINVAL);          // Invalid argument.
      return (-1);
    }
  //
  // The beginning of the packet contains the IP header.
  //
  header = (struct iphdr *) packet;
  //
  // Verify IP header checksum: it is also calculated the real header
  // size.
  //
  size_header = header->ihl * 4;
  checksum    = ip_checksum ((uint16_t *) header, size_header,
                             NULL, (size_t) 0);
  if (checksum == 0xFFFF || checksum == 0x0000)
    {
      ; // k_printf ("checksum ok\n");
    }
  else
    {
      k_printf ("BAD CHECKSUM: %04x\n", checksum);
      return (0);
    }
  //
  // Is it a fragment? As we are not able to manage fragments,
  // we just check that it is all zero, ignoring the bit `DF'.
  // That is why we use a mask 0xBFFF.
  //
  if ((ntohs (header->frag_off) & 0xBFFF) != 0)
    {
      //
      // Sorry, we don't manage fragments.
      //
      k_printf ("Sorry: we don't manage IP fragments: %04x\n",
                (ntohs (header->frag_off)));
      return (0);
    }
  else
    {
      //
      // Find a place inside the IP table.
      //
      ip_table_item = ip_reference ();
      //
      // Copy the packet inside the ip_table[] item, then update
      // the link to the data start inside the packet and the
      // clock_t timestamp.
      //
      memcpy (ip_table_item->packet.octet, packet,
              ntohs (header->tot_len));
      ip_table_item->pdu4  = ip_table_item->packet.octet;
      ip_table_item->pdu4 += (header->ihl * 4);
      ip_table_item->clock = k_clock ();
    }
  //
  // Check for destination unreachable, scanning the interface table,
  // to see if there is such address here.
  //
  for (j = 0; j < NET_MAX_DEVICES; j++)
    {
      if (net_table[j].ip == ntohl (header->daddr))
        {
          //
          // Found a valid local address.
          //
          break;
        }
    }
  if (j >= NET_MAX_DEVICES)
    {
      //
      // Local address not found: host unreachable, but the packet
      // is taken anyway.
      //
      icmp_tx_unreachable (ntohl (header->daddr),
                           ntohl (header->saddr),
                           ICMP_DEST_UNREACH, ICMP_HOST_UNREACH,
                           packet->octet,
                           ntohs (header->tot_len));
    }
  //
  // Check for port unreachable, scanning the socket table.
  //
  if (header->protocol == IPPROTO_UDP
      || header->protocol == IPPROTO_TCP)
    {
      //
      // There are ports.
      //
      udp = (struct udphdr *) &(packet->octet[header->ihl*4]);
      //
      for (s = 0; s < SOCK_MAX_SLOTS; s++)
        {
          if (sock_table[s].active
              && sock_table[s].lport == ntohs (udp->dest))
            {
              //
              // Found a matching local port.
              //
              break;
            }
        }
      if (s >= SOCK_MAX_SLOTS)
        {
          //
          // Local port not found: port unreachable, but the packet
          // is taken anyway.
          //
          icmp_tx_unreachable (ntohl (header->daddr),
                               ntohl (header->saddr),
                               ICMP_DEST_UNREACH, ICMP_PORT_UNREACH,
                               packet->octet,
                               ntohs (header->tot_len));
        }
    }
  //
  // Now do something with the data inside the `ip_table[]'.
  //
  for (i = 0; i < IP_MAX_PACKETS; i++)
    {
      if (ip_table[i].clock != 0)
        {
          //
          if (ip_table[i].kernel_serviced != ip_table[i].clock
              && ip_table[i].packet.header.protocol == IPPROTO_ICMP)
            {
              icmp_rx (i);
            }
          else
            {
              //
              // At the moment, no other IP protocol managed internally.
              //
              ip_table[i].kernel_serviced = ip_table[i].clock;
            }
        }
    }
  //
  //
  //
  return (0);
}
