#include <kernel/net.h>
#include <kernel/net/ip.h>
#include <kernel/net/route.h>
#include <sys/os32.h>
#include <kernel/lib_k.h>
#include <errno.h>
#include <arpa/inet.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
ip_tx (h_addr_t src, h_addr_t dst, int protocol, const void *buffer,
       size_t size)
{
  static int       id = 0;
  ip_packet_t      packet;
  uint16_t         checksum;
  int              s;         // source net interface.
  int              d;         // destination net interface.
  net_buffer_lo_t *loopback;
  //
  // Verify to have a source address.
  //
  if (src == 0)
    {
      //
      // Default source address: get the source address from the routing
      // table, based on the destination.
      //
      src = route_remote_to_local (dst);
      if (src == ((h_addr_t) -1))
        {
          errset (errno);
          return (-1);
        }
    }
  //
  // Prepare the packet.
  //
  packet.header.version  = IP_VERSION;
  packet.header.ihl      = sizeof (struct iphdr) / 4;
  packet.header.tos      = 0x0000;                  // Routine, normal.
  packet.header.tot_len  = htons (sizeof (struct iphdr) + size);
  packet.header.id       = htons (id++);
  packet.header.frag_off = htons (0x4000);           // Do not fragment!
  packet.header.ttl      = IP_TTL;
  packet.header.protocol = protocol;
  packet.header.check    = 0;
  packet.header.saddr    = htonl (src);
  packet.header.daddr    = htonl (dst);
  //
  // Now set the header checksum.
  //
  checksum = ~(ip_checksum ((void *) &packet, sizeof (struct iphdr),
                            NULL, (size_t) 0));
  packet.header.check = htons (checksum);
  //
  memcpy (packet.data, buffer, size);
  //
  //////////////////////////////////////////////////////////////////////
  // Enter here the lower network level.
  //////////////////////////////////////////////////////////////////////
  //
  // The new size includes now the IPv4 header
  //
  size = (sizeof (struct iphdr) + size);
  //
  // Check for PDU size.
  //
  if (size > NET_MTU)
    {
      errset (E_PDU_TOO_BIG);
      return (-1);
    }
  //
  // Find the sender interface.
  //
  s = net_index (src);
  if (s < 0)
    {
      errset (errno);           // ENODEV.
      return (-1);
    }
  //
  // Check if the destination is a local interface.
  //
  d = net_index (dst);
  if (d >= 0)
    {
      //
      // It is a local interface, so must change the destination
      // to the loopback device: it must be `net0'.
      //
      d = 0;
    }
  else
    {
      //
      // Should not be necessary, but for coherence with the rest
      // of the code...
      //
      d = s;
    }
  //
  // Check if the destination is the loopback interface.
  //
  if (net_table[d].type & NET_DEV_LOOP)
    {
      loopback = net_buffer_lo (d);
      if (loopback == NULL)
        {
          errset (errno);
          return (-1);
        }
      loopback->clock = k_clock ();
      loopback->size  = size;
      memcpy (&loopback->packet, (void *) &packet, size);
      return (0);
    }
  //
  // The destination wasn't the loopback interface, so check if the
  // source is an Ethernet device.
  //
  if (net_table[s].type & NET_DEV_ETH)
    {
      //
      // For Ethernet devices another function is responsible
      // for sending the packet.
      //
      return (net_eth_ip_tx (src, dst, (void *) &packet, size));
    }
  //
  // Should never reach the end, but who knows...
  //
  errset (errno);           // ENODEV.
  return (-1);
}
