#include <kernel/net.h>
#include <kernel/net/ip.h>
#include <kernel/net/route.h>
#include <kernel/net/tcp.h>
#include <sys/os32.h>
#include <kernel/lib_k.h>
#include <kernel/fs.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
tcp_tx_sock (void *sock_item)
{
  sock_t             *sock = sock_item;
  tcp_packet_t        packet;
  tcp_pseudo_header_t pseudo;
  uint16_t            checksum;
  size_t              tcp_size;
  uint32_t            seq;
  uint32_t            ack_seq;
  size_t              send_size;
  //
  if (sock == NULL)
    {
      errset (EINVAL);
      return (-1);
    }
  if (!sock->tcp.can_send)
    {
      errset (EINVAL);
      return (-1);
    }
  if (sock->laddr == 0 || sock->raddr == 0)
    {
      errset (EINVAL);
      return (-1);
    }
  if (sock->lport == 0 || sock->rport == 0)
    {
      errset (EINVAL);
      return (-1);
    }
  //
  // Sequences and size.
  //
  seq = sock->tcp.lsq[sock->tcp.lsqi];
  if ((sock->tcp.send_flags & TCP_FLAG_SYN)
      || (sock->tcp.send_flags & TCP_FLAG_FIN))
    {
      //
      // A SYN or FIN packet cannot load data.
      //
      send_size = 0;
      //
      // The next expected ACK from the other side is just
      // +1.
      //
      sock->tcp.lsq_ack = seq + 1;
    }
  else
    {
      send_size = sock->tcp.send_size;
      //
      // The next expected ACK from the other side is
      // + size of the sent data.
      //
      sock->tcp.lsq_ack = seq + send_size;
    }
  //
  if (sock->tcp.send_flags & TCP_FLAG_ACK)
    {
      ack_seq = sock->tcp.rsq[sock->tcp.rsqi];
    }
  else
    {
      ack_seq = 0;
    }
  //
  // Prepare the TCP packet.
  //
  memset (&packet.header, 0, sizeof (struct tcphdr));
  //
  packet.header.source   = htons (sock->lport);
  packet.header.dest     = htons (sock->rport);
  packet.header.seq      = htonl (seq);
  packet.header.ack_seq  = htonl (ack_seq);
  packet.header.doff     = (sizeof (struct tcphdr) / 4);        // 5
  if (sock->tcp.send_flags & TCP_FLAG_ACK) packet.header.ack = 1;
  if (sock->tcp.send_flags & TCP_FLAG_PSH) packet.header.psh = 1;
  if (sock->tcp.send_flags & TCP_FLAG_RST) packet.header.rst = 1;
  if (sock->tcp.send_flags & TCP_FLAG_SYN) packet.header.syn = 1;
  if (sock->tcp.send_flags & TCP_FLAG_FIN) packet.header.fin = 1;
  if (sock->tcp.send_flags & TCP_FLAG_RST)
    {
      packet.header.window = htons (0);
    }
  else
    {
      packet.header.window = htons (TCP_MSS);        // Minimal window
    }
  packet.header.check    = 0;
  //
  memcpy (packet.data, sock->tcp.send_data, send_size);
  //
  tcp_size = sizeof (struct tcphdr) + send_size;
  //
  // Prepare the pseudo header.
  //
  pseudo.saddr    = htonl (sock->laddr);
  pseudo.daddr    = htonl (sock->raddr);
  pseudo.zero     = 0;
  pseudo.protocol = IPPROTO_TCP;
  pseudo.length   = htons (tcp_size);
  //
  // Now set the header checksum.
  //
  checksum = ~(ip_checksum ((void *) &packet, tcp_size,
                            (void *) &pseudo,
                            sizeof (tcp_pseudo_header_t)));
  if (checksum == 0)
    {
      checksum = 0xFFFF;
    }
  packet.header.check = htons (checksum);
  //
  // Send to the lower network level.
  //
  if (DEBUG)
    {
      tcp_show (sock->laddr, sock->raddr, (struct tcphdr *) &packet);
    }
  sock->tcp.clock = s_clock ((pid_t) 0);
  return (ip_tx (sock->laddr, sock->raddr, (int) IPPROTO_TCP,
                 &packet, tcp_size));
}
