#include <kernel/proc.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
#include <kernel/driver/kbd.h>
#include <termios.h>
#include <limits.h>
//----------------------------------------------------------------------
void
proc_sch_terminals (void)
{
  pid_t         pid;
  pid_t         pid_sub;
  unsigned char key;
  tty_t        *tty;
  dev_t         device;
  int           k;            // Reverse count killed character.
  int           overflow = 0; // True if input is too much.
  //
  // Try to read a key from console keyboard buffer (only consoles
  // are available). Variable `kbd' is external and comes from
  // `kernel/kbd.h'.
  //
  key = kbd.key;
  if (key == 0)
    {
      //
      // No key is ready on the keyboard buffer: just return.
      //
      return;
    }
  else
    {
      //
      // A key was pressed and it will be processed.
      // Currently, just remove from the keyboard buffer.
      kbd.key  = 0;
    }
  //
  // A key is available. Find the currently active console.
  //
  device = tty_console ((dev_t) 0);
  tty    = tty_reference (device);
  if (tty == NULL)
    {
      k_printf ("kernel alert: console device 0x%04x not found!\n",
                device);
      //
      // Will send the typed character to the first terminal!
      //
      tty = tty_reference ((dev_t) 0);
    }
  //
  // Verify if it is a control key that must be handled before
  // entering the canonical input line.
  //
  // Check for a console switch key combination.
  //
  if      (key == 0x11)               // [Ctrl Q] -> DC1 -> console0.
    {
      tty_console (DEV_CONSOLE0);     // Switch.
      return;
    }
  else if (key == 0x12)               // [Ctrl R] -> DC2 -> console1.
    {
      tty_console (DEV_CONSOLE1);     // Switch.
      return;
    }
  else if (key == 0x13)               // [Ctrl S] -> DC3 -> console2.
    {
      tty_console (DEV_CONSOLE2);     // Switch.
      return;
    }
  else if (key == 0x14)               // [Ctrl T] -> DC4 -> console3.
    {
      tty_console (DEV_CONSOLE3);     // Switch.
      return;
    }
  //------------------------------------------------------------------
  // Only canonical input line is available: ICANON!
  //------------------------------------------------------------------
  //
  // Might be necessary to send a signal to all processes with the
  // same control terminal, excluded the kernel (0) and `init' (1).
  // Such control keys are not passed to the applications, or are
  // replaced with zero.
  //
  // Please note that this a simplified solution, because the signal
  // should reach only the foreground process of the group. For that
  // reason, only che [Ctrl C] is taken into consideration, because
  // processes can ignore the signal 'SIGINT'.
  //
  if (tty->pgrp != 0)
    {
      //
      // There is a process group for that terminal.
      //
      if (tty->attr.c_cc[VINTR] && key == tty->attr.c_cc[VINTR])
        {
          if (tty->attr.c_iflag & IGNBRK)
            {
              //
              // No break!
              //
              return;
            }
          //
          if (tty->attr.c_iflag & BRKINT)
            {
              if (tty->attr.c_lflag & ISIG)
                {
                  for (pid = 2; pid <  PROCESS_MAX; pid++)
                    {
                      if (proc_table[pid].pgrp == tty->pgrp)
                        {
                          //
                          // Should find only final process/processes.
                          // So, if there is a son process with the
                          // same process group, the signal is not
                          // sent!
                          //
                          for (pid_sub = 2;
                               pid_sub < PROCESS_MAX;
                               pid_sub++)
                            {
                                if ((proc_table[pid_sub].ppid == pid)
                                    && (proc_table[pid_sub].pgrp
                                        == tty->pgrp))
                                  {
                                    //
                                    // `pid_sub' is a son of
                                    // `pid' and is part of the
                                    // same process group. `pid_sub'
                                    // is candidate for kill.
                                    //
                                    break;
                                  }
                             }
                          //
                          if (pid_sub >= PROCESS_MAX)
                            {
                              //
                              // There is no son for `pid', sorry.
                              //
                              s_kill ((pid_t) 0, pid, SIGINT);
                              //
                              // No more scan.
                              //
                              break;
                            }
                        }
                    }
                }
              //
              // Just reset input line and return.
              //
              tty->status  = TTY_INPUT_LINE_EDITING;
              tty->lpr     = 0;
              tty->lpw     = 0;
              tty->line[0] = 0;
              //
              return;
            }
          //
          // Replace the INTR character with zero.
          //
          key = 0;
        }
    }
  //
  // Check if something is to be ignored.
  //
  if (key == '\r' && (tty->attr.c_iflag & IGNCR))
    {
      return;
    }
  //
  // Check if something is to be replaced, before editing.
  //
  if (key == '\n' && (tty->attr.c_iflag & INLCR))
    {
      key = '\r';
    }
  //
  if (key == '\r' && (tty->attr.c_iflag & ICRNL))
    {
      key = '\n';
    }
  //
  // Edit the canonical input line.
  // Input is accepted only if status is ok.
  //
  if (tty->status == TTY_INPUT_LINE_EDITING)
    {
      //
      // Fix internal line positions.
      //
      if (tty->lpw < 0)
        {
          tty->lpw = 0;
        }
      if (tty->lpw >= MAX_CANON)
        {
          tty->lpw = MAX_CANON-1;
          overflow = 1;                 // Too much input!
        }
      if (tty->lpr < 0)
        {
          tty->lpr = 0;
        }
      if (tty->lpr > tty->lpw)
        {
          tty->lpr = tty->lpw;
        }
      //
      // Check input key.
      //
      if (key == '\0')
        {
          //
          // A INTR replaced into zero.
          //
          tty->line[tty->lpw] = key;
          tty->status = TTY_INPUT_LINE_CLOSED;
          proc_wakeup_terminal ();
        }
      else if (key == '\n')
        {
          tty->line[tty->lpw] = key;
          tty->status = TTY_INPUT_LINE_CLOSED;
          proc_wakeup_terminal ();
        }
      else if (tty->attr.c_cc[VEOF] && key == tty->attr.c_cc[VEOF])
        {
          //
          // EOF is not included inside the line: just replace the
          // with zero.
          //
          key = 0;
          tty->line[tty->lpw] = key;
          tty->status = TTY_INPUT_LINE_CLOSED;
          proc_wakeup_terminal ();
        }
      else if (tty->attr.c_cc[VEOL] && key == tty->attr.c_cc[VEOL])
        {
          tty->line[tty->lpw] = key;
          tty->status = TTY_INPUT_LINE_CLOSED;
          proc_wakeup_terminal ();
        }
      else if (   tty->attr.c_cc[VERASE]
               && key == tty->attr.c_cc[VERASE])
        {
          //
          // Save how many characters to be killed, if echo is
          // enabled.
          //
          if (overflow)
            {
              k = tty->lpw + 1;
              //
              // The `tty->lpw' was already reduced.
              //
            }
          else
            {
              k = tty->lpw;
              //
              // Reduce write index: if it is less than zero, it will
              // be fixed.
              //
              tty->lpw--;
            }
        }
      else if (tty->attr.c_cc[VKILL] && key == tty->attr.c_cc[VKILL])
        {
          //
          // Save how many characters to be killed, if echo is
          // enabled.
          //
          if (overflow)
            {
              k = tty->lpw + 1;
            }
          else
            {
              k = tty->lpw;
            }
          //
          // Reset input line.
          //
          tty->lpw     = 0;
          tty->lpr     = 0;
          tty->line[0] = 0;
        }
      else
        {
          tty->line[tty->lpw] = key;
          tty->lpw++;
        }
      //
      // Echo.
      //
      if (!(tty->attr.c_lflag & ECHO))
        {
          //
          // No echo. But NL might be echoed anyway.
          //
          if (key == '\n' && (tty->attr.c_lflag & ECHONL))
            {
              dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                      (off_t) 0, &key, (size_t) 1, NULL);
            }
          //
          return;
        }
      //
      // The echo is requested.
      //
      if (key == 0)
        {
          //
          // There is nothing to echo.
          //
          ;
        }
      else if (   tty->attr.c_cc[VERASE]
               && key == tty->attr.c_cc[VERASE])
        {
          if (tty->attr.c_lflag & ECHOE)
            {
              if (k > 0)
                {
                  dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                          (off_t) 0, "\b \b", (size_t) 3, NULL);
                }
            }
          else
            {
              dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                      (off_t) 0, &key, (size_t) 1, NULL);
            }
       }
      else if (   tty->attr.c_cc[VKILL]
               && key == tty->attr.c_cc[VKILL])
        {
          if (tty->attr.c_lflag & ECHOK)
            {
              for (; k > 0; k--)
                {
                  dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                          (off_t) 0, "\b \b", (size_t) 3, NULL);
                }
            }
        }
      else if (key == '\n')
        {
          if (tty->attr.c_lflag & ECHONL)
            {
              dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                      (off_t) 0, &key, (size_t) 1, NULL);
            }
        }
      else
        {
          //
          // If there was an overflow, the last character is
          // overwriting the last position, so a back space
          // is printed.
          //
          if (overflow)
            {
              dev_io ((pid_t) 0, tty->device, DEV_WRITE,
                      (off_t) 0, "\b", (size_t) 1, NULL);
            }
          //
          // Now show the input character.
          //
          dev_io ((pid_t) 0, tty->device, DEV_WRITE, (off_t) 0,
                  &key, (size_t) 1, NULL);
        }
    }
}
