#include <kernel/proc.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
#include <stdint.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
void
proc_sig_handler (pid_t pid, int sig)
{
    addr_t   addr_data_top;
    addr_t   addr_stack_pointer;
    uint32_t old_eip;
    uint32_t old_cs;
    uint32_t old_eflags;
    //
    // Stack frames.
    //
    struct {uint32_t eax;
            uint32_t ecx;
            uint32_t edx;
            uint32_t ebx;
            uint32_t ebp;
            uint32_t esi;
            uint32_t edi;
            uint32_t ds;
            uint32_t es;
            uint32_t fs;
            uint32_t gs;
            uint32_t eip;
            uint32_t cs;
            uint32_t eflags;} *old;
    //
    struct {uint32_t eax;
            uint32_t ecx;
            uint32_t edx;
            uint32_t ebx;
            uint32_t ebp;
            uint32_t esi;
            uint32_t edi;
            uint32_t ds;
            uint32_t es;
            uint32_t fs;
            uint32_t gs;
            uint32_t wrapper;
            uint32_t cs;
            uint32_t eflags;
            uint32_t handler;
            uint32_t signal;
            uint32_t eip;}    *new;
    //
    // First of all, this function can act only for a process that
    // is not *just* interrupted. That is, if `proc_current' is
    // equal to `pid', nothing is to be done yet: it will be done
    // only when it will be in pause.
    //
    if (pid == proc_current)
      {
        //
        // Nothing to do yet.
        //
        return;
      }
    //
    // Check if there is a function to run.
    //
    if (proc_table[pid].sig_handler[sig] == (uintptr_t) NULL)
      {
        //
        // Nothing to do.
        //
        return;
      }
    //
    // Tell something for debugging.
    //
    if (DEBUG)
      {
        k_printf ("%s(%i, %i)", __func__, (int) pid, sig);
      }
    //
    // Calculate the absolute stack section address, from the
    // kernel point of view.
    //
    if (proc_table[pid].domain_data == 0)
      {
        addr_data_top = proc_table[pid].address_text;
      }
    else
      {
        addr_data_top = proc_table[pid].address_data;
      }
    //
    // Then calculate the effective stack pointer address.
    // We are considering only process that are not
    // currently interrupted, so the stack pointer is
    // taken from `proc_table[pid].sp'.
    //
    addr_stack_pointer = addr_data_top + proc_table[pid].sp;
    //
    // Currently the process stack is as it follows.
    // The address inside `addr_stack_pointer' is corresponding
    // to the saved EAX value.
    //
    //  pushl %eflags     #
    //  pushl %cs         # from the interrupt
    //  pushl %eip        #
    //  -------------
    //  pushl %gs
    //  pushl %fs
    //  pushl %es
    //  pushl %ds
    //  pushl %edi
    //  pushl %esi
    //  pushl %ebp
    //  pushl %ebx
    //  pushl %edx
    //  pushl %ecx
    //  pushl %eax
    //
    // Need to insert the call to the handler function.
    // The process stack should become this way:
    //
    //  pushl %eip            [0]
    //  pushl <signal>        [1]
    //  pushl <handler>       [1]
    //  pushl %eflags         [2]
    //  pushl %cs             [2]
    //  pushl <wrapper>       [2]
    //  -------------
    //  pushl %gs
    //  pushl %fs
    //  pushl %es
    //  pushl %ds
    //  pushl %edi
    //  pushl %esi
    //  pushl %ebp
    //  pushl %ebx
    //  pushl %edx
    //  pushl %ecx
    //  pushl %eax
    //
    // [0] Back from the original interrupt.
    //
    // [1] Arguments of the wrapper functions, that will call the
    //     signal handler, and then will return to the address at
    //     [1].
    //
    // [2] Modified so that the IRET instruction will return
    //     to the begin of the wrapper function, that will
    //     call the true signal handler, and will return
    //     at [1].
    //
    // Now set the pointer to the old and new frame,
    // updating the stack pointer address.
    //
    old                 = (void *) addr_stack_pointer;
    addr_stack_pointer -= 12;    // Three more elements.
    new                 = (void *) addr_stack_pointer;
    //
    // Verify if the code segment was correctly found.
    //
    if (DEBUG)
      {
        k_printf ("[%s] EAX:%04x ECX:%04x EDX:%04x EBX:%04x EBP:%04x "
                  "ESI:%04x EDI:%04x DS:%04x ES:%04x FS:%04x GS:%04x "
                  "EIP:%04x CS:%04x EFLAGS:%04x\n", __FILE__,
                  (int) old->eax, (int) old->ecx, (int) old->edx,
                  (int) old->ebx, (int) old->ebp, (int) old->esi,
                  (int) old->edi, (int) old->ds,  (int) old->es,
                  (int) old->fs,  (int) old->gs,  (int) old->eip,
                  (int) old->cs,  (int) old->eflags);
      }
    //
    // Move data, to arrange the new stack. The order does
    // matter, as the new frame overwrites the old one.
    //
    new->eax = old->eax;
    new->ecx = old->ecx;
    new->edx = old->edx;
    new->ebx = old->ebx;
    new->ebp = old->ebp;
    new->esi = old->esi;
    new->edi = old->edi;
    new->ds  = old->ds;
    new->es  = old->es;
    new->fs  = old->fs;
    new->gs  = old->gs;
    //
    old_eflags = old->eflags;
    old_cs     = old->cs;
    old_eip    = old->eip;
    //
    new->wrapper = proc_table[pid].sig_handler_wrapper;
    new->cs      = old_cs;
    new->eflags  = old_eflags;
    new->handler = proc_table[pid].sig_handler[sig];
    new->signal  = sig;
    new->eip     = old_eip;
    //
    // Tell what is changed inside the stack.
    //
    if (DEBUG)
      {
        k_printf ("[%s] EAX:%04x ECX:%04x EDX:%04x EBX:%04x EBP:%04x "
                  "ESI:%04x EDI:%04x DS:%04x ES:%04x FS:%04x GS:%04x "
                  "wrapper:%04x CS:%04x EFLAGS:%04x handler:%04x "
                  "signal:$04x EIP:%04x\n", __FILE__,
                  (int) new->eax, (int) new->ecx, (int) new->edx,
                  (int) new->ebx, (int) new->ebp, (int) new->esi,
                  (int) new->edi, (int) new->ds,  (int) new->es,
                  (int) new->fs,  (int) new->gs,  (int) new->wrapper,
                  (int) new->cs,  (int) new->eflags,
                  (int) new->handler, (int) new->signal,
                  (int) new->eip);
      }
    //
    // Change the stack pointer of the process, as it was increased.
    //
    proc_table[pid].sp = addr_stack_pointer - addr_data_top;
    //
    // Reset the signal handler, as in traditional Unix, with
    // all the consequences that such implementation will give.
    //
    proc_table[pid].sig_handler[sig] = (uintptr_t) NULL;
    proc_sig_off (pid, sig);
    //
    // Wake up the process if it is sleeping.
    //
    proc_table[pid].wakeup_events = 0;
    proc_table[pid].status        = PROC_READY;
}
